% W-H system simulation program
% nonlinear system y = av1*x + av2*x^2 + av3*x^3 + av4*x^4+ av5*x^5,
clear all; close all; rand('seed',0);
%
NLG=0.5;
av=[1 NLG*[0.1 0.2 0.05 0.02]];

Plevel=1; Plevel2=Plevel^2; 
% excitation signals: 'irg' - ideal gaussian, 'rms' - multisine, 'rmso1' - odd rms,

ML=100;                  %number of measurements
npoints=1024*2*4; 		% number of time domain points
n=npoints/2;     		% maximum number of frequencies
nfreq=fix(npoints/10);
junit=sqrt(-1);
%
% frequency axes
fv=[1:nfreq]';
fvo1=[1:2:nfreq]'; fveven=[2:2:nfreq]';
fvo2=[1:4:nfreq]';
%
% W-H system design, 'L' input dynamics, 'R' output dynamics
[b1,a1]=butter(3,0.03,'high'); 
[b2,a2]=butter(3,0.07);

fvrel=0.1*fv/nfreq; zet=exp(-sqrt(-1)*2*pi*fvrel);
HLf=(b1(1)+b1(2)*zet+b1(3)*zet.^2+b1(4)*zet.^3)./(a1(1)+a1(2)*zet+a1(3)*zet.^2+a1(4)*zet.^3);
HRf=(b2(1)+b2(2)*zet+b2(3)*zet.^2+b2(4)*zet.^3)./(a2(1)+a2(2)*zet+a2(3)*zet.^2+a2(4)*zet.^3);
LRf=HLf.*HRf;
%
% BLA gain
RDS1=1; RDS3=3*sum(abs(HLf(fv)).^2)/length(fv);
RDS5=5*3*(sum(abs(HLf(fv)).^2)/length(fv))^2;
RDStr=av(1)*RDS1+Plevel2*av(3)*RDS3+Plevel2^2*av(5)*RDS5;

RDS1o1=1; RDS3o1=3*sum(abs(HLf(fvo1)).^2)/length(fvo1);
RDS5o1=5*3*(sum(abs(HLf(fvo1)).^2)/length(fvo1))^2;
RDStro1=av(1)*RDS1o1+Plevel2*av(3)*RDS3o1+Plevel2^2*av(5)*RDS5o1;

RDS1o2=1; RDS3o2=3*sum(abs(HLf(fvo2)).^2)/length(fvo2);
RDS5o2=5*3*(sum(abs(HLf(fvo2)).^2)/length(fvo2))^2;
RDStro2=av(1)*RDS1o2+Plevel2*av(3)*RDS3o2+Plevel2^2*av(5)*RDS5o2;
%
figure(1),
 subplot(2,2,1), plot(fvrel,db(HLf)); grid; axis([0 .1 -50 10]);
 title('Input System','Fontsize',10);
 subplot(2,2,2), plot(fvrel,db(HRf)); grid; axis([0 .1 -50 10]);
 title('Output System','Fontsize',10);
 subplot(2,2,4), plot(fvrel,db(LRf)); grid; axis([0 .1 -50 10]);
 title('Linear System','Fontsize',10);
%
Xrms=zeros(npoints,1); Xrmso1=Xrms; Xirg=Xrms; Xrmso2=Xrms;
X1rms=Xrms; X1rmso1=Xrms; X1irg=Xrms; X1rmso2=Xrms;
X2rms=Xrms; X2rmso1=Xrms; X2irg=Xrms; X2rmso2=Xrms;
% measured system characteristics
Grms=zeros(length(fv),1); Grmso1=zeros(length(fvo1),1); 
Grmso2=zeros(length(fvo2),1); Girg=zeros(length(fv),1); 
Syxirg=Girg; Sxxirg=Girg; 
GrmsAv=0*Grms; Grmso1Av=0*Grmso1; Grmso2Av=0*Grmso2; GirgAv=0*Girg; 
Veven=zeros(length(fveven),1);
%
for ll=1:ML, ll
% xrms, Xrms - random phase multisine
 fi=rand(length(fv),1)*2*pi; Zx=exp(-junit*fi); Xrms=0*Xrms;
 Xrms(fv+1)=Zx; Xrms(2*n+1-fv)=conj(Xrms(1+fv));
 xrms=real(ifft(Xrms)); Ax=sqrt(sum(xrms.^2)/npoints);  xrms=xrms/Ax;
 xrms=Plevel*xrms;
 Xrms=fft(xrms); AW=abs(Xrms)/max(abs(Xrms));

% xrmso1, Xrmso1 - random phase odd freq multisine
 fio1=rand(length(fvo1),1)*2*pi; Zxo1=exp(-junit*fio1);
 Xrmso1=0*Xrmso1;
 Xrmso1(fvo1+1)=Zxo1; Xrmso1(2*n+1-fvo1)=conj(Xrmso1(1+fvo1));
 xrmso1=real(ifft(Xrmso1)); Ax=sqrt(sum(xrmso1.^2)/npoints); xrmso1=xrmso1/Ax;
 xrmso1=Plevel*xrmso1;
 Xrmso1=fft(xrmso1); 

% xrmso2, Xrmso2 - random phase odd-odd freq multisine
 fio2=rand(length(fvo2),1)*2*pi; Zxo2=exp(-junit*fio2);
 Xrmso2=0*Xrmso2;
 Xrmso2(fvo2+1)=Zxo2; Xrmso2(2*n+1-fvo2)=conj(Xrmso2(1+fvo2));
 xrmso2=real(ifft(Xrmso2)); Ax=sqrt(sum(xrmso2.^2)/npoints); xrmso2=xrmso2/Ax;
 xrmso2=Plevel*xrmso2;
 Xrmso2=fft(xrmso2); 

% xirg, Xirg - ideal filtered Gaussian signal
 xaux=randn(length(xrms),1); Xaux=fft(xaux); Xirg=Xaux.*AW;
 xirg=real(ifft(Xirg)); Ax=sqrt(sum(xirg.^2)/npoints); xirg=xirg/Ax;
 xirg=Plevel*xirg;
 Xirg=fft(xirg);
%
% passing signals through input dynamics
X1rms(fv+1)=Xrms(fv+1).*HLf;
X1rms(2*n+1-fv)=conj(X1rms(1+fv));
x1rms=real(ifft(X1rms));

X1rmso1(fvo1+1)=Xrmso1(fvo1+1).*HLf(fvo1);
X1rmso1(2*n+1-fvo1)=conj(X1rmso1(1+fvo1));
x1rmso1=real(ifft(X1rmso1));

X1rmso2(fvo2+1)=Xrmso2(fvo2+1).*HLf(fvo2);
X1rmso2(2*n+1-fvo2)=conj(X1rmso2(1+fvo2));
x1rmso2=real(ifft(X1rmso2));

X1irg(fv+1)=Xirg(fv+1).*HLf;
X1irg(2*n+1-fv)=conj(X1irg(1+fv));
x1irg=real(ifft(X1irg));

% passing signals through nonlinear system
x2rms=0*x1rms; x2rmso1=0*x1rms; x2irg=0*x1rms; x2rmso2=0*x1rms; 

if av(1)~=0,
  x2rms=x2rms+av(1)*x1rms; x2rmso1=x2rmso1+av(1)*x1rmso1;
  x2irg=x2irg+av(1)*x1irg; x2rmso2=x2rmso2+av(1)*x1rmso2;
end;
if av(2)~=0,
  x2rms=x2rms+av(2)*x1rms.^2; x2rmso1=x2rmso1+av(2)*x1rmso1.^2;
  x2irg=x2irg+av(2)*x1irg.^2;  x2rmso2=x2rmso2+av(2)*x1rmso2.^2;
end;
if av(3)~=0,
  x2rms=x2rms+av(3)*x1rms.^3; x2rmso1=x2rmso1+av(3)*x1rmso1.^3;
  x2irg=x2irg+av(3)*x1irg.^3; x2rmso2=x2rmso2+av(3)*x1rmso2.^3;
end;
if av(4)~=0,
  x2rms=x2rms+av(4)*x1rms.^4; x2rmso1=x2rmso1+av(4)*x1rmso1.^4;
  x2irg=x2irg+av(4)*x1irg.^4; x2rmso2=x2rmso2+av(4)*x1rmso2.^4;
end;
if av(5)~=0,
  x2rms=x2rms+av(5)*x1rms.^5; x2rmso1=x2rmso1+av(5)*x1rmso1.^5;
  x2irg=x2irg+av(5)*x1irg.^5; x2rmso2=x2rmso2+av(5)*x1rmso2.^5;
end;
%
% passing signals through output dynamics
 X2rms=fft(x2rms); X2rmsB=X2rms(fv+1);
 X2rmso1=fft(x2rmso1); X2rmso1B=X2rmso1(fvo1+1);
 X2rmso2=fft(x2rmso2); X2rmso2B=X2rmso2(fvo2+1);
 X2irg=fft(x2irg); X2irgB=X2irg(fv+1);

 YrmsBR=X2rmsB.*HRf;
 Yrmso1BR=X2rmso1B.*HRf(fvo1);
 Yrmso2BR=X2rmso2B.*HRf(fvo2);
 YirgBR=X2irgB.*HRf;
 
% computing estimates of the system transfer characteristics
 Grms=YrmsBR./Xrms(fv+1);
 Grmso1=Yrmso1BR./Xrmso1(fvo1+1);
 Grmso2=Yrmso2BR./Xrmso2(fvo2+1);
 
 GrmsAv=GrmsAv+(Grms-GrmsAv)/ll;
 Grmso1Av=Grmso1Av+(Grmso1-Grmso1Av)/ll;
 Grmso2Av=Grmso2Av+(Grmso2-Grmso2Av)/ll;
 
 Veven=Veven+(abs(Xrms(fveven+1)).^2-Veven)/ll;

% for gaussian signals
 Syxirg=YirgBR.*conj(Xirg(fv+1));
 Sxxirg=Xirg(fv+1).*conj(Xirg(fv+1));
 Girg=Syxirg./Sxxirg;
 
 GirgAv=GirgAv+(Girg-GirgAv)/ll;

 figure(4),
 plot(fv,db(LRf),'r',fv,db(RDStr*LRf),'g',fv,db(Grms(:)),'k',fv,db(GrmsAv(:)),'b');
 axis([0 1000 -30 10]); grid;
 title('r - lin, g - BLAlin, b - rms, k - rms ave');
 drawnow;
end;

figure(3),
subplot(2,2,2), plot(db(Xrms)); xlabel('Xrms','Fontsize',10);
subplot(2,2,3), plot(max(db(Xrmso1),-300)); xlabel('Xrmso1','Fontsize',10);
subplot(2,2,4), plot(max(db(Xrmso2),-300)); xlabel('Xrmso2','Fontsize',10);
subplot(2,2,1), plot(db(Xirg)); xlabel('Xirg','Fontsize',10);
drawnow;

yrmsout=filter(b2,a2,x2rms);
Yrms=fft(yrmsout);

figure(8), 
subplot(2,2,1), plot(xrms); axis([0 length(xrms) -10 10]);
grid; title('Input','Fontsize',10);
subplot(2,2,3), plot(yrmsout); axis([0 length(yrmsout) -10 10]);
grid; title('Output','Fontsize',10);
subplot(2,2,2), plot(db(Xrms),'LineWidth',1.5);
grid; title('Input','Fontsize',10);
axis([0 2000 -50 90]);
subplot(2,2,4), plot(db(Yrms));
grid; title('Output','Fontsize',10);
axis([0 2000 -50 90]);
%
xv=[-3:.01:3];
yv=av(1)*xv+av(2)*xv.^2+av(3)*xv.^3+av(4)*xv.^4+av(5)*xv.^5;

figure(10), 
 subplot(1,4,1), plot(fvrel,db(HLf),'LineWidth',2); grid; axis([0 .1 -50 10]);
 title('Input System','Fontsize',10);
 subplot(1,4,2), plot(xv,yv,'LineWidth',2); grid;
 title('Nonlinearity','Fontsize',10);
 subplot(1,4,3), plot(fvrel,db(HRf),'LineWidth',2); grid; axis([0 .1 -50 10]);
 title('Output System','Fontsize',10);
 subplot(1,4,4), plot(fvrel,db(LRf),'LineWidth',2); grid; axis([0 .1 -50 10]);
 title('Linear System','Fontsize',10);

figure(11),
 subplot(1,4,1), plot(fvrel,db(Girg(:)),'k');
 ylabel('                           G(f)                                  dB','Fontsize',10);
 axis([0 .1 -30 10]); grid; title('Gauss noise','Fontsize',10);
 subplot(1,4,2), plot(fvrel,db(Grms(:)),'k');
 axis([0 .1 -30 10]); grid; title('Random phase MS','Fontsize',10);
 subplot(1,4,3), plot(fvrel(fvo1),db(Grmso1(:)),'k');
 axis([0 .1 -30 10]); grid; title('odd RPhMS','Fontsize',10);
 subplot(1,4,4), plot(fvrel(fvo2),db(Grmso2(:)),'k');
 xlabel('                 Relative freq','Fontsize',10);
 axis([0 .1 -30 10]); grid; title('odd-odd RPhMS','Fontsize',10);

figure(12),
 subplot(1,4,1), plot(fvrel,db(GirgAv(:)),'b');
 ylabel('                           G(f)                                  dB','Fontsize',10);
 axis([0 .1 -30 10]); grid; title('Gauss noise','Fontsize',10);
 subplot(1,4,2), plot(fvrel,db(GrmsAv(:)),'b');
 axis([0 .1 -30 10]); grid; title('Random phase MS','Fontsize',10);
 subplot(1,4,3), plot(fvrel(fvo1),db(Grmso1Av(:)),'b');
 axis([0 .1 -30 10]); grid; title('odd RPhMS','Fontsize',10);
 subplot(1,4,4), plot(fvrel(fvo2),db(Grmso2Av(:)),'b');
 xlabel('                 Relative freq','Fontsize',10);
 axis([0 .1 -30 10]); grid; title('odd-odd RPhMS','Fontsize',10);
 

