% NAME: imtc2004-mimo4.m
% SIMO with 4 inputs to experiment with general U, WxU(l), HxU(l), and
% [WU], [HU] inputs
% same grids
clear all; close all; rand('seed',0);

NLG = .5;
%------------------------------------
%
a1=1; a2=1; a3=1; a4=1;
aorder1=[a1, a2, a3, a4];
%
a11=.5; a12=.2; a13=.2; a14=.2; a22=.5; a23=.2; a24=.2; a33=.5; a34=.2; a44=.5;
aorder2=NLG*[a11, a12, a13, a14, a22, a23, a24, a33, a34, a44];
%
a111=.2; a112=.2; a113=.2; a114=.2; a122=.2; a123=.2; a124=.2; a133=.2; a134=.2; a144=.2; 
a222=.2; a223=.2; a224=.2; a233=.2; a234=.2; a244=.2; a333=.2; a334=.2; a344=.2; a444=.2; 
aorder3=NLG*[a111, a112, a113, a114, a122, a123, a124, a133, a134, a144, a222, a223, a224, a233, ...
        a234, a244, a333, a334, a344, a444]; 
%
% excitation signals: odd multisine
%--------------------------------------
junit=-sqrt(-1); 
ML=10;  					%number of measurements
%
npoints1=1024*2*8; 		    % number of time domain points
n1=npoints1/2;     			% maximum number of frequencies
nfreq1=fix(npoints1/10); 
fv1=[1:nfreq1]'; 
fvo1=[1:2:nfreq1]'; %fvo1=fv1;
NN1=2*length(fvo1);
%
% Fourier matrix
w11=1; w12=1; w13=1; w14=1;
w21=1; w22=exp(junit*2*pi/4); w23=exp(junit*4*pi/4); w24=exp(junit*6*pi/4);
w31=1; w32=exp(junit*4*pi/4); w33=exp(junit*8*pi/4); w34=exp(junit*12*pi/4);
w41=1; w42=exp(junit*6*pi/4); w43=exp(junit*12*pi/4); w44=exp(junit*18*pi/4);
%
% Hadamard matrix
w11=1; w12=1; w13=1; w14=1;
w21=1; w22=-1; w23=1; w24=-1;
w31=1; w32=1; w33=-1; w34=-1;
w41=1; w42=-1; w43=-1; w44=1;

orNR1=9; [br1,ar1]=cheby1(orNR1,25,0.15); 
orNR2=3; [br2,ar2]=cheby1(orNR2,10,0.02,'high'); 
orNR3=9; [br3,ar3]=cheby1(orNR3,10,0.15); 
orNR4=9; [br4,ar4]=cheby1(orNR4,25,0.05,'high'); 
orNS1=3; [bs1,as1]=cheby1(orNS1,1,0.07);
%
fvrel1=0.1*fv1/nfreq1; zet1=exp(-sqrt(-1)*2*pi*fvrel1);
%
HR1fD=br1(1); HR1fN=ar1(1); 
for or=1:orNR1, HR1fD=HR1fD+br1(or+1)*zet1.^or; HR1fN=HR1fN+ar1(or+1)*zet1.^or; end; 
R1f1=(HR1fD./HR1fN);
%
HR2fD=br2(1); HR2fN=ar2(1); 
for or=1:orNR2, HR2fD=HR2fD+br2(or+1)*zet1.^or; HR2fN=HR2fN+ar2(or+1)*zet1.^or; end; 
R2f1=(HR2fD./HR2fN);
%
HR3fD=br3(1); HR3fN=ar3(1); 
for or=1:orNR3, HR3fD=HR3fD+br3(or+1)*zet1.^or; HR3fN=HR3fN+ar3(or+1)*zet1.^or; end; 
R3f1=(HR3fD./HR3fN);
%
HR4fD=br4(1); HR4fN=ar4(1); 
for or=1:orNR4, HR4fD=HR4fD+br4(or+1)*zet1.^or; HR4fN=HR4fN+ar4(or+1)*zet1.^or; end; 
R4f1=(HR4fD./HR4fN);
%
HS1fD=bs1(1); HS1fN=as1(1); 
for or=1:orNS1, HS1fD=HS1fD+bs1(or+1)*zet1.^or; HS1fN=HS1fN+as1(or+1)*zet1.^or; end;
S1f1=(HS1fD./HS1fN);
%
R1S1f1=R1f1.*S1f1; 
R2S1f1=R2f1.*S1f1; 
R3S1f1=R3f1.*S1f1; 
R4S1f1=R4f1.*S1f1; 

% BLA dynamic systems
PR1f1p2=sum(abs(R1f1(fvo1)).^2)/length(fvo1); 
PR2f1p2=sum(abs(R2f1(fvo1)).^2)/length(fvo1); 
PR3f1p2=sum(abs(R3f1(fvo1)).^2)/length(fvo1); 
PR4f1p2=sum(abs(R4f1(fvo1)).^2)/length(fvo1); 
%
RDR1S1full=a1+NLG*(a111*3*PR1f1p2+a122*PR2f1p2+a133*PR3f1p2+a144*PR4f1p2);
RDR2S1full=a2+NLG*(a222*3*PR2f1p2+a112*PR1f1p2+a233*PR3f1p2+a244*PR4f1p2);
RDR3S1full=a3+NLG*(a333*3*PR3f1p2+a113*PR1f1p2+a223*PR2f1p2+a344*PR4f1p2);
RDR4S1full=a4+NLG*(a444*3*PR4f1p2+a114*PR1f1p2+a224*PR2f1p2+a334*PR3f1p2);
%
figure(1),
 subplot(4,3,1), plot(fvrel1,db(R1f1)); grid; title('R1 Input System','FontName','Arial','Fontsize',12);
 subplot(4,3,2), plot(fvrel1,db(S1f1)); grid; title('S1 Output System','FontName','Arial','Fontsize',12);
 subplot(4,3,3), plot(fvrel1,db(R1S1f1)); grid; title('R1*S1 Linear System','FontName','Arial','Fontsize',12);
%
 subplot(4,3,4), plot(fvrel1,db(R2f1)); grid; title('R2 Input System','FontName','Arial','Fontsize',12);
 subplot(4,3,5), plot(fvrel1,db(S1f1)); grid; title('S1 Output System','FontName','Arial','Fontsize',12);
 subplot(4,3,6), plot(fvrel1,db(R2S1f1)); grid; title('R2*S1 Linear System','FontName','Arial','Fontsize',12);
%
 subplot(4,3,7), plot(fvrel1,db(R3f1)); grid; title('R3 Input System','FontName','Arial','Fontsize',12);
 subplot(4,3,8), plot(fvrel1,db(S1f1)); grid; title('S1 Output System','FontName','Arial','Fontsize',12);
 subplot(4,3,9), plot(fvrel1,db(R3S1f1)); grid; title('R3*S1 Linear System','FontName','Arial','Fontsize',12);
%
 subplot(4,3,10), plot(fvrel1,db(R4f1)); grid; title('R4 Input System','FontName','Arial','Fontsize',12);
 subplot(4,3,11), plot(fvrel1,db(S1f1)); grid; title('S1 Output System','FontName','Arial','Fontsize',12);
 subplot(4,3,12), plot(fvrel1,db(R4S1f1)); grid; title('R4*S1 Linear System','FontName','Arial','Fontsize',12);
%
X1=zeros(npoints1,1); Xo1=X1; Zo1=X1; Vo1=X1; Yo1=zeros(length(fv1),1); 
X2=zeros(npoints1,1); Xo2=X2;
X3=zeros(npoints1,1); Xo3=X3;
X4=zeros(npoints1,1); Xo4=X4;
%
Go11m=zeros(length(fvo1),1); Go12m=zeros(length(fvo1),1); Go13m=zeros(length(fvo1),1); Go14m=zeros(length(fvo1),1); 
Go11t=zeros(length(fvo1),1); Go12t=zeros(length(fvo1),1); Go13t=zeros(length(fvo1),1); Go14t=zeros(length(fvo1),1); 
%
Nc1=0;
while Nc1<ML, % generation of excitation signals
%
% 1st experiment
fio1=rand(length(fvo1),1)*2*pi; Zxo1=exp(-junit*fio1); Xo1=0*Xo1; Xo1(fvo1+1)=Zxo1; 
Xo1(2*n1+1-fvo1)=conj(Xo1(1+fvo1)); xo1=npoints1*real(ifft(Xo1)); Ax1=sqrt(sum(xo1.^2)/npoints1); 
xo1ref=xo1/Ax1; Xo1ref=Xo1/Ax1;
%
fio1=rand(length(fvo1),1)*2*pi; Zxo1=exp(-junit*fio1); Xo2=0*Xo2; Xo2(fvo1+1)=Zxo1; 
Xo2(2*n1+1-fvo1)=conj(Xo2(1+fvo1)); xo2=npoints1*real(ifft(Xo2)); Ax2=sqrt(sum(xo2.^2)/npoints1); 
xo2ref=xo2/Ax2; Xo2ref=Xo2/Ax2;
%
fio1=rand(length(fvo1),1)*2*pi; Zxo1=exp(-junit*fio1); Xo3=0*Xo3; Xo3(fvo1+1)=Zxo1; 
Xo3(2*n1+1-fvo1)=conj(Xo3(1+fvo1)); xo3=npoints1*real(ifft(Xo3)); Ax3=sqrt(sum(xo3.^2)/npoints1); 
xo3ref=xo3/Ax3; Xo3ref=Xo3/Ax3;
%
fio1=rand(length(fvo1),1)*2*pi; Zxo1=exp(-junit*fio1); Xo4=0*Xo4; Xo4(fvo1+1)=Zxo1; 
Xo4(2*n1+1-fvo1)=conj(Xo4(1+fvo1)); xo4=npoints1*real(ifft(Xo4)); Ax4=sqrt(sum(xo4.^2)/npoints1); 
xo4ref=xo4/Ax4; Xo4ref=Xo4/Ax4;
%
% 1st experiment
xo1=xo1ref; Xo1=Xo1ref; 
xo2=xo2ref; Xo2=Xo2ref; 
xo3=xo3ref; Xo3=Xo3ref; 
xo4=xo4ref; Xo4=Xo4ref;
%
% passing signals through input dynamics
Zo1(fvo1+1)=Xo1(fvo1+1).*R1f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo1=npoints1*real(ifft(Zo1));
Zo1(fvo1+1)=Xo2(fvo1+1).*R2f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo2=npoints1*real(ifft(Zo1));
Zo1(fvo1+1)=Xo3(fvo1+1).*R3f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo3=npoints1*real(ifft(Zo1));
Zo1(fvo1+1)=Xo4(fvo1+1).*R4f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo4=npoints1*real(ifft(Zo1));
% passing signals through nonlinear system
vo1=passingNLmimo4(zo1, zo2, zo3, zo4, aorder1, aorder2, aorder3);
% passing signals through output dynamics
Ybe1=fft(vo1)/npoints1; YS1=Ybe1(fv1+1); Yo11e1=YS1.*S1f1(fv1);
%
% 2nd experiment
xo1=xo1ref; Xo1=Xo1ref; 
xo2=-xo2ref; Xo2=-Xo2ref; 
xo3=xo3ref; Xo3=Xo3ref; 
xo4=-xo4ref; Xo4=-Xo4ref;
%
% passing signals through input dynamics
Zo1(fvo1+1)=Xo1(fvo1+1).*R1f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo1=npoints1*real(ifft(Zo1));
Zo1(fvo1+1)=Xo2(fvo1+1).*R2f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo2=npoints1*real(ifft(Zo1));
Zo1(fvo1+1)=Xo3(fvo1+1).*R3f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo3=npoints1*real(ifft(Zo1));
Zo1(fvo1+1)=Xo4(fvo1+1).*R4f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo4=npoints1*real(ifft(Zo1));
% passing signals through nonlinear system
vo1=passingNLmimo4(zo1, zo2, zo3, zo4, aorder1, aorder2, aorder3);
% passing signals through output dynamics
Ybe1=fft(vo1)/npoints1; YS1=Ybe1(fv1+1); Yo11e2=YS1.*S1f1(fv1);
%
% 3rd experiment
xo1=xo1ref; Xo1=Xo1ref; 
xo2=xo2ref; Xo2=Xo2ref; 
xo3=-xo3ref; Xo3=-Xo3ref; 
xo4=-xo4ref; Xo4=-Xo4ref;
%
% passing signals through input dynamics
Zo1(fvo1+1)=Xo1(fvo1+1).*R1f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo1=npoints1*real(ifft(Zo1));
Zo1(fvo1+1)=Xo2(fvo1+1).*R2f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo2=npoints1*real(ifft(Zo1));
Zo1(fvo1+1)=Xo3(fvo1+1).*R3f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo3=npoints1*real(ifft(Zo1));
Zo1(fvo1+1)=Xo4(fvo1+1).*R4f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo4=npoints1*real(ifft(Zo1));
% passing signals through nonlinear system
vo1=passingNLmimo4(zo1, zo2, zo3, zo4, aorder1, aorder2, aorder3);
% passing signals through output dynamics
Ybe1=fft(vo1)/npoints1; YS1=Ybe1(fv1+1); Yo11e3=YS1.*S1f1(fv1);
%
% 4th experiment
xo1=xo1ref; Xo1=Xo1ref; 
xo2=-xo2ref; Xo2=-Xo2ref; 
xo3=-xo3ref; Xo3=-Xo3ref; 
xo4=xo4ref; Xo4=Xo4ref;
%
% passing signals through input dynamics
Zo1(fvo1+1)=Xo1(fvo1+1).*R1f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo1=npoints1*real(ifft(Zo1));
Zo1(fvo1+1)=Xo2(fvo1+1).*R2f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo2=npoints1*real(ifft(Zo1));
Zo1(fvo1+1)=Xo3(fvo1+1).*R3f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo3=npoints1*real(ifft(Zo1));
Zo1(fvo1+1)=Xo4(fvo1+1).*R4f1(fvo1); Zo1(2*n1+1-fvo1)=conj(Zo1(1+fvo1)); zo4=npoints1*real(ifft(Zo1));
% passing signals through nonlinear system
vo1=passingNLmimo4(zo1, zo2, zo3, zo4, aorder1, aorder2, aorder3);
% passing signals through output dynamics
Ybe1=fft(vo1)/npoints1; YS1=Ybe1(fv1+1); Yo11e4=YS1.*S1f1(fv1);
%
Ye1=Yo11e1(fvo1); Ye2=Yo11e2(fvo1); Ye3=Yo11e3(fvo1); Ye4=Yo11e4(fvo1); 
X1e1=Xo1ref(fvo1+1); X2e1=Xo2ref(fvo1+1); X3e1=Xo3ref(fvo1+1); X4e1=Xo4ref(fvo1+1); 
%
% set of equations, result Go11t, Go12t for frequency 1 ... N 
Gain=1/4/abs(X1e1(1))^2;
%
for k=1:length(fvo1),
    Yvec=[Ye1(k), Ye2(k), Ye3(k), Ye4(k)].';
    Xmat=[X1e1(k),  X2e1(k),  X3e1(k),  X4e1(k); 
          X1e1(k), -X2e1(k),  X3e1(k), -X4e1(k);
          X1e1(k),  X2e1(k), -X3e1(k), -X4e1(k); 
          X1e1(k), -X2e1(k), -X3e1(k),  X4e1(k);];
%    Gvec=inv(Xmat)*Yvec;
    Gvec=Gain*Xmat'*Yvec;
    Go11t(k)=Gvec(1); Go12t(k)=Gvec(2); Go13t(k)=Gvec(3); Go14t(k)=Gvec(4);
end;
%
Nc1=Nc1+1,
%
Go11m(:)=Go11m(:)+(Go11t-Go11m(:))/Nc1; Go12m(:)=Go12m(:)+(Go12t-Go12m(:))/Nc1;
Go13m(:)=Go13m(:)+(Go13t-Go13m(:))/Nc1; Go14m(:)=Go14m(:)+(Go14t-Go14m(:))/Nc1;
%
figure(2), 
subplot(221), plot(fvo1,db(a1*R1S1f1(fvo1)),'g',fvo1,db(RDR1S1full*R1S1f1(fvo1)),'m',fvo1,db(Go11t),'k',fvo1,db(Go11m),'b');
title('Systems in channel 1','FontName','Arial','Fontsize',14); grid; drawnow;
subplot(222), plot(fvo1,db(a2*R2S1f1(fvo1)),'g',fvo1,db(RDR2S1full*R2S1f1(fvo1)),'m',fvo1,db(Go12t),'k',fvo1,db(Go12m),'b');
title('Systems in channel 2','FontName','Arial','Fontsize',14); grid; drawnow;
subplot(223), plot(fvo1,db(a3*R3S1f1(fvo1)),'g',fvo1,db(RDR3S1full*R3S1f1(fvo1)),'m',fvo1,db(Go13t),'k',fvo1,db(Go13m),'b');
title('Systems in channel 3','FontName','Arial','Fontsize',14); grid; drawnow;
subplot(224), plot(fvo1,db(a4*R4S1f1(fvo1)),'g',fvo1,db(RDR4S1full*R4S1f1(fvo1)),'m',fvo1,db(Go14t),'k',fvo1,db(Go14m),'b');
title('Systems in channel 4','FontName','Arial','Fontsize',14); grid; drawnow;
%
end;
%
figure(20), 
subplot(221), plot(fvo1,db(Go11m),'b',fvo1,db(a1*R1S1f1(fvo1)),'g',fvo1,db(Go11t),'k'); grid; axis([0 1700 -100 10]);
subplot(222), plot(fvo1,db(Go12m),'b',fvo1,db(a2*R2S1f1(fvo1)),'g',fvo1,db(Go12t),'k'); grid; axis([0 1700 -100 10]);
subplot(223), plot(fvo1,db(Go13m),'b',fvo1,db(a3*R3S1f1(fvo1)),'g',fvo1,db(Go13t),'k'); grid; axis([0 1700 -100 10]);
subplot(224), plot(fvo1,db(Go14m),'b',fvo1,db(a4*R4S1f1(fvo1)),'g',fvo1,db(Go14t),'k'); grid; axis([0 1700 -100 10]);

